#!/usr/bin/perl

# findextract.pl - A script which grabs Essence properties and material stocks from
#                  the ATITD wiki and generates up to the 20 "cheapest" possible 
#                  Compound Extract recipes that match properties given at the 
#                  command line.

# Required for the 'get' command
use LWP::Simple;

# Parse the command-line options
my $numreqs;
if( ($numreqs = @ARGV) == 0 ) {
   print "Usage: Run " . $0 . " with the properties you're looking for in an Extract.\n";
   print "Example:\n";
   print "   " . $0 . " Ar++++ So----\n";
   exit;
}
my @reqtype, @reqvalue;
for( my $i=0 ; $i<@ARGV ; $i++ ) {
   ($reqtype[$i], $reqvalue[$i]) = $ARGV[$i] =~ /(Ar|As|Bi|Sa|So|Sp|Sw|To)([+-]*)/;
   $reqvalue[$i] = ((3*length($reqvalue[$i]))-2) * ("," cmp $reqvalue[$i]);
}

# The address of the page from which to draw Essence values.
my $essencepage = get("http://www.atitd.org/wiki/tale4/Chemistry/Essences");
# The address of the page from which to draw essential material stocks.
my $stockpage = get("http://www.atitd.org/wiki/tale4/User:Numaris/Notes");

# Variable declarations/initializations
my @stocklines = split(/\n/, $stockpage);
my @essencelines = split(/<tr>/, $essencepage);
@essencelines = @essencelines[2..($#essencelines-1)];
my @essencename = ();
my @essencestats = ();
my @essencevalue = ();
my %index = ( Ar => 0, As => 1, Bi => 2, Sa => 3, So => 4, Sp => 5, Sw => 6, To => 7 );
my @toptwenty = ( [30000,,,,,], [30000,,,,,], [30000,,,,,], [30000,,,,,], [30000,,,,,],
                  [30000,,,,,], [30000,,,,,], [30000,,,,,], [30000,,,,,], [30000,,,,,],
                  [30000,,,,,], [30000,,,,,], [30000,,,,,], [30000,,,,,], [30000,,,,,],
                  [30000,,,,,], [30000,,,,,], [30000,,,,,], [30000,,,,,], [30000,,,,,] );
my @currstat1, @currstat2, @currstat3, @currstat4, @currstat5;
my $cost;

# Parse each line of the Essence table and load values into @essencename and @essencestats
for( my $i=0 ; $i<@essencelines ; $i++ ) {
   $essencelines[$i] =~ s/<\/?a.*?>//g;
   $essencelines[$i] =~ s/<\/?s.*?>//g;
   ($essencename[$i], my $dummy, my @v) = $essencelines[$i] =~ /<td>[\s]*([^<\s][^<]*[^<\s])[\s]*(<\/td><td>[^<]*){3}<\/td><td>[\s]+([^<\s]*)[\s]*<\/td><td>[\s]+([^<\s]*)[\s]*<\/td><td>[\s]+([^<\s]*)[\s]*<\/td><td>[\s]+([^<\s]*)[\s]*<\/td><td>[\s]+([^<\s]*)[\s]*<\/td><td>[\s]+([^<\s]*)[\s]*<\/td><td>[\s]+([^<\s]*)[\s]*<\/td><td>[\s]+([^<\s]*)[\s]*.*/;
   $essencestats[$i] = [@v];
}

# Remove Essences whose attributes have not yet been tested
for( my $i=0 ; $i<@essencename ; $i++ ) {
   if( $essencestats[$i][0] eq "" ) {
      splice( @essencename, $i, 1 );
      splice( @essencestats, $i, 1 );
      $i--;
   }
}

# Determine the values of the essences based on the available stock. This section
# is sort of complicated. I can't guarantee that what works for me will work for
# you, but I also can't really go into too much detail about what's going on here. 
foreach $essence ( 0 .. $#essencename ) {
   if( my ($gemtype) = ($essencename[$essence] =~ /Powdered (.*)/) ) {
      if( $gemtype =~ /(Quartz|Ruby|Topaz|Emerald|Sapphire|Diamond)/ ) {
          # Value of mined gem powders
          $essencevalue[$essence] = 150;
      } else {
          # Value of cuttable gem powders
          $essencevalue[$essence] = 400;
      }
   } elsif( $essencename[$essence] =~ /Salts of/ ) {
      # Value of metal salts
      $essencevalue[$essence] = 300;
   } else {
      my $found = 0;
      foreach( @stocklines ) {
         if( ($qty) = /(\d+) Resin:$essencename[$essence]/ ) {
            # Value of resins, based on available stock
            $essencevalue[$essence] = ( 50 / ( ($qty / 40) + 1) );
            $found++;
            last;
         } elsif( ($qty) = /(\d+) $essencename[$essence]/ ) {
            # Value of herbs (and anything else), based on available stock
            $essencevalue[$essence] = ( 1000 / ( ($qty / 30) + 1 ) );
            $found++;
            last;
         }
      }
      if( $found == 0 ) {
         # Value of anything not found above, which at the present includes
         # Camel Pheromones, Marble Dust, and anything I'm out of stock on.
         # This section serves two purposes: to give a high default value to
         # anything I don't stock, and to give a heads up if something on
         # the wiki might be misspelled.
         $essencevalue[$essence] = 5000;
         print "No match: " . $essencename[$essence] . "\n";
      }
   }
}

# Start looping through all the available essences to find a combination of five that satisfies
# the given requirements. This is a fairly brute-force method, though it does try to be intelligent
# and move quickly past any combinations that can't possibly work. As a result, the time to run
# this command will grow exponentially as the Essence list grows, but the tighter requirements
# (Ra, Thoth, etc) should still run pretty quickly.
ONE: for( my $i=0 ; $i<(@essencename-4) ; $i++ ) {
   foreach( 0..7 ) {
      $currstat1[$_] = $essencestats[$i][$_];
   }
   for( my $x=0 ; $x<$numreqs ; $x++ ) {
      if( ((($reqvalue[$x] <=> 0)*$currstat1[$index{$reqtype[$x]}]) + 12) < abs($reqvalue[$x]) ){
         next ONE;
      }
   }

   TWO: for( my $j=$i+1 ; $j<(@essencename-3) ; $j++ ) {
      foreach( 0..7 ) {
         $currstat2[$_] = $currstat1[$_] + $essencestats[$j][$_];
      }
      for( my $x=0 ; $x<$numreqs ; $x++ ) {
         if( ((($reqvalue[$x] <=> 0)*$currstat2[$index{$reqtype[$x]}]) + 9) < abs($reqvalue[$x]) ){
            next TWO;
         }
      }

      THREE: for( my $k=$j+1 ; $k<(@essencename-2) ; $k++ ) {
         foreach( 0..7 ) {
            $currstat3[$_] = $currstat2[$_] + $essencestats[$k][$_];
         }
         for( my $x=0 ; $x<$numreqs ; $x++ ) {
            if( ((($reqvalue[$x] <=> 0)*$currstat3[$index{$reqtype[$x]}]) + 6) < abs($reqvalue[$x]) ){
               next THREE;
            }
         }

         FOUR: for( my $l=$k+1 ; $l<(@essencename-1) ; $l++ ) {
            foreach( 0..7 ) {
               $currstat4[$_] = $currstat3[$_] + $essencestats[$l][$_];
            }
            for( my $x=0 ; $x<$numreqs ; $x++ ) {
               if( ((($reqvalue[$x] <=> 0)*$currstat4[$index{$reqtype[$x]}]) + 3) < abs($reqvalue[$x]) ){
                  next FOUR;
               }
            }

            FIVE: for( my $m=$l+1 ; $m<@essencename ; $m++ ) {
               foreach( 0..7 ) {
                  $currstat5[$_] = $currstat4[$_] + $essencestats[$m][$_];
               }
               for( my $x=0 ; $x<$numreqs ; $x++ ) {
                  if( (($reqvalue[$x] <=> 0)*$currstat5[$index{$reqtype[$x]}]) < abs($reqvalue[$x]) ){
                     next FIVE;
                  }
               }

               # At this point, we've found a viable Compound Extract. Sum the values of the essences
               # and sort it into the Top 20.
               $cost = $essencevalue[$i] + $essencevalue[$j] + $essencevalue[$k] + $essencevalue[$l] + $essencevalue[$m];
               for( my $y=19 ; $y>=0 ; $y-- ) {
                  if( $cost > $toptwenty[$y][0] ) {
                     splice( @toptwenty,$y+1,0,[$cost,$essencename[$i],$essencename[$j],$essencename[$k],$essencename[$l],$essencename[$m]] );
                     last;
                  } elsif( $y == 0 ) {
                     unshift( @toptwenty,[$cost,$essencename[$i],$essencename[$j],$essencename[$k],$essencename[$l],$essencename[$m]] );
                  }
               }
               pop( @toptwenty );


            }
         }
      }
   }
}

# Print out the results
foreach( 0 .. 19 ) {
   if( $toptwenty[$_][0] < 30000 ) {
       print $toptwenty[$_][1] . ", " . $toptwenty[$_][2] . ", " . $toptwenty[$_][3] . ", " . $toptwenty[$_][4] . ", " . $toptwenty[$_][5] . ": " . int($toptwenty[$_][0]) . "\n";
   }
}
